/****************************************************************************/
//
// maplistmgmt.cpp
//
// OVERVIEW
// ~~~~~~~~
// maplistmgmt.cpp is used to maintain a list of numbers using the STL map.
// This module can be instaniated for each list of numbers.  The
// conversions for the numbers to different number bases is included.
//
/****************************************************************************/
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#pragma hdrstop

#include "maplistmgmt.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

//---------------------------------------------------------------------------
// This function is the constructor for the class.  The constructor accepts
// an integer parameter as input to preset the number base the list should be
// in.  This function also clears the filtered and unfiltered lists.
//---------------------------------------------------------------------------
MapListMgmt::MapListMgmt(int mode)
{
   // constructor
   int i;

   // set the number base according to what the user wants
   switch (mode) {
      case NUM_MGMT_DECIMAL:
      case NUM_MGMT_OCTAL:
      case NUM_MGMT_HEXADECIMAL:
         NumberBase = mode;
         break;
      default:  // default to decimal if a valid number was not entered
         NumberBase = NUM_MGMT_DECIMAL;
   }

   // ensure all the lists are clear
   for (i=0; i<NUM_MGMT_MAX_LIST; i++)
      NUM_Mgmt_List[i].clear();
}

//---------------------------------------------------------------------------
// This function is the destructor of the class.  This function ensures the
// lists are cleared and the memory used is freed.
//---------------------------------------------------------------------------
MapListMgmt::~MapListMgmt()
{
   // destructor
   int i;

   // clear all the lists
   for (i=0; i<NUM_MGMT_MAX_LIST; i++)
      NUM_Mgmt_List[i].clear();
}

//---------------------------------------------------------------------------
// Function DecimalToHexadecimal converts an integer number from base 10 to
// a hexadecimal string.
//---------------------------------------------------------------------------
void MapListMgmt::DecimalToHexadecimal(int DecimalIn, char *HexOutput)
{
   sprintf(HexOutput, "%.4X", DecimalIn);
}

//---------------------------------------------------------------------------
// Function OctalToDecimal converts a string containing an octal number to a
// decimal number and returns the integer.
//---------------------------------------------------------------------------
int MapListMgmt::OctalToDecimal(char *OctalIn)
{
   int returnvalue;

   sscanf(OctalIn, "%o", &returnvalue);

   return (returnvalue);
}

//---------------------------------------------------------------------------
// Function HexadecimalToDecimal converts a string containing a hexadecimal
// number to a decimal number and returns the integer.
//---------------------------------------------------------------------------
int MapListMgmt::HexadecimalToDecimal(char *HexadecimalIn)
{
   int returnvalue;

   sscanf(HexadecimalIn, "%x", &returnvalue);

   return (returnvalue);
}

//---------------------------------------------------------------------------
// Function DecimalToDecimal converts a string containing a decimal number
// to an integer and returns the integer.
//---------------------------------------------------------------------------
int MapListMgmt::DecimalToDecimal(char *DecimalIn)
{
   int returnvalue;

   sscanf(DecimalIn, "%d", &returnvalue);

   return (returnvalue);
}

//---------------------------------------------------------------------------
// Function DecimalToOctal converts a decimal integer to a charcter string that
// represents the number in octal format.
//---------------------------------------------------------------------------
void MapListMgmt::DecimalToOctal(int DecimalIn, char *OctalOutput)
{
   sprintf(OctalOutput, "%.4o", DecimalIn);
}

//---------------------------------------------------------------------------
// Function ChangeToInteger takes the character string and converts it to a
// decimal integer.  The conversion is determined by the number base of the
// lists.  This function is used when a number is being added to or removed from
// a list.  In these situations the number has to be an integer in order
// to see whether it is in the list or not.
//---------------------------------------------------------------------------
int MapListMgmt::ChangeToInteger(char *ThisNumber)
{
   int returnvalue=0;

   switch (NumberBase) {
      case NUM_MGMT_DECIMAL:
         returnvalue = DecimalToDecimal(ThisNumber);
         break;
      case NUM_MGMT_OCTAL:
         returnvalue = OctalToDecimal(ThisNumber);
         break;
      case NUM_MGMT_HEXADECIMAL:
         returnvalue = HexadecimalToDecimal(ThisNumber);
         break;
   }

   return (returnvalue);
}

//---------------------------------------------------------------------------
// Function CopyList copies a map list to either the filtered or unfiltered map
// list.  The parameter Use_List determines whether the data will be copied to
// the filtered or unfiltered list.  This function is used when a list from
// another instantiation of this class is to be used to set the lists in this
// instantiation of the class.
//---------------------------------------------------------------------------
void MapListMgmt::CopyList(NUM_Mgmt_List_Type ToBeCopied, int Use_List)
{
   NUM_Mgmt_Iterator_Type iter;
   int i, numitems;

   if (!ToBeCopied.empty())  {  // if there is data in the list

      // get the number of elements to copy
      numitems = ToBeCopied.size();

      // set the iterator to the first item
      iter = ToBeCopied.begin();

      if (Use_List == NUM_MGMT_FILTERED_LIST) {

         // clear the destination list
         NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].clear();

         // copy the list item by item
         for (i=0; i<numitems; i++) {
            // insert the item into the new list
            NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].insert(NUM_Mgmt_Entry(iter->first,
                                                                   iter->second));
            // increment the iterator to the next item
            iter++;
         }

      }
      else {

         // clear the destination list
         NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].clear();

         // copy the list item by item
         for (i=0; i<numitems; i++) {
            // insert the item into the new list
            NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].insert(NUM_Mgmt_Entry(iter->first,
                                                                   iter->second));
            // increment the iterator to the next item
            iter++;
         }

      }

   }

}

//---------------------------------------------------------------------------
// Function ChangeMode changes the number base of the lists.  The parameter
// newbase indicates the number base to change to.  Use_List indicates which
// list (filtered or unfiltered) will be copied into the TListBox.
//---------------------------------------------------------------------------
void MapListMgmt::ChangeMode(TListBox *ListToUpdate, int newbase, int Use_List)
{
   NUM_Mgmt_Iterator_Type iter, iterend;
   int i, numitems;

   // change from one base to another
   switch (newbase) {
      case NUM_MGMT_DECIMAL:
      case NUM_MGMT_OCTAL:
      case NUM_MGMT_HEXADECIMAL:
         NumberBase = newbase;
         break;
      default:  // default to decimal if a valid number was not entered
         NumberBase = NUM_MGMT_DECIMAL;
   }

   // check to make sure there is a TListBox and not NULL
   if (ListToUpdate) {

      // clear out the listbox before copying the items into it
      ListToUpdate->Clear();

      // set the iterators to the first and last items
      if (Use_List == NUM_MGMT_FILTERED_LIST) {

         iter = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].begin();
         if (!NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].empty()) {
            iterend = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].end();
         }
         else {
            iterend = iter;
         }

      }
      else {

         iter = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].begin();
         if (!NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].empty()) {
            iterend = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].end();
         }
         else {
            iterend = iter;
         }

      }

      // copy the list item by item
      while (iter != iterend) {

         // switch according to the base the user wants
         switch (NumberBase) {
            case NUM_MGMT_DECIMAL:
               // insert the item into the new list
               ListToUpdate->Items->Append(iter->second.Decimal);
               break;
            case NUM_MGMT_OCTAL:
               // insert the item into the new list
               ListToUpdate->Items->Append(iter->second.Octal);
               break;
            case NUM_MGMT_HEXADECIMAL:
               // insert the item into the new list
               ListToUpdate->Items->Append(iter->second.Hexadecimal);
               break;
         }
         // increment the iterator to the next item
         iter++;

      }

   }

}

//---------------------------------------------------------------------------
// Function AddNewNumber adds the number to either the filtered or
// unfiltered list.  The function first checks to see if the number is
// already in one of the lists.  If the number is not in either list then
// the number will be added to the indicated list.
//---------------------------------------------------------------------------
void MapListMgmt::AddNewNumber(int DecimalNumber, bool filter) {
   int check;                  // check will indicate if number exists in the list
   NumberFormats NumberStrings;

   check = FilterStatus(DecimalNumber);

   // if the number is not in either list then add it
   if (check == NUM_MGMT_NOT_IN_LIST) {

      // convert the number to the different strings
      sprintf(NumberStrings.Decimal, "%d", DecimalNumber);
      DecimalToOctal(DecimalNumber, NumberStrings.Octal);
      DecimalToHexadecimal(DecimalNumber, NumberStrings.Hexadecimal);

      if (filter) {  // put it in the filtered list
         NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].insert(NUM_Mgmt_Entry(DecimalNumber,
                                                                NumberStrings));
      }
      else {  // put it in the unfiltered list
         NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].insert(NUM_Mgmt_Entry(DecimalNumber,
                                                                NumberStrings));
      }

   }

}

//---------------------------------------------------------------------------
// Function RemoveNumber will remove the desired number from the lists.
//---------------------------------------------------------------------------
int MapListMgmt::RemoveNumber(char *ThisNumber)
{
   int DecimalNumber;   // the integer representation of the input string
   int fstatus;     // will indicate whether the number is in a list

   // convert the string to an integer
   DecimalNumber = ChangeToInteger(ThisNumber);
   // see which list the number is in
   fstatus = FilterStatus(DecimalNumber);

   switch (fstatus) {
      case NUM_MGMT_NOT_IN_LIST:  // not in either list
         // do nothing
         break;
      case NUM_MGMT_UNFILTERED:   // in the unfiltered list
         NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].erase(DecimalNumber);
         break;
      case NUM_MGMT_FILTERED:     // in the filtered list
         NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].erase(DecimalNumber);
         break;
   }

   // return which list the number was removed from
   return (fstatus);
}

//---------------------------------------------------------------------------
// Function AddToFilter takes the selected items from the TListBox and adds
// them to the filtered list.  The items will not be added to the filtered list
// unless the number is in the unfiltered list.  So this function cannot
// be used to add a new number to the filtered list.
//---------------------------------------------------------------------------
void MapListMgmt::AddToFilter(TListBox *UnfilteredList)
{
   int numcount;
   int DecimalNumber, i;
   NUM_Mgmt_Iterator_Type iter;

   // get the number of items in the TListBox
   numcount = UnfilteredList->Items->Count;

   // loop through each item in the TListBox
   for (i=0; i < numcount; i++) {

      // if the current item is selected then process it
      if (UnfilteredList->Selected[i]) {

         // change the string to a decimal integer
         DecimalNumber = ChangeToInteger(UnfilteredList->Items->Strings[i].c_str());
         // look for the number in the unfiltered list
         iter = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].find(DecimalNumber);

         // if the number was found in the unfiltered list
         if (iter != NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].end()) {

            // copy the item to the filtered list
            NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].insert(NUM_Mgmt_Entry(iter->first,
                                                                  iter->second));
            // delete the item from the unfiltered list
            NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].erase(iter);
         }

         // delete the item from the TListBox
         UnfilteredList->Items->Delete(i);
         // decrement the number of items because we deleted one
         numcount--;
         // decrement the loop variable so we do not skip over an item
         i--;
      }  // end if (UnfilteredList->Selected[i])

   }  // end of the for loop

}

//---------------------------------------------------------------------------
// Function AddToUnfilter takes the selected items from the TListBox and adds
// them to the unfiltered list.  The items will not be added to the unfiltered
// list unless the number is in the filtered list.  So this function cannot
// be used to add a new number to the unfiltered list.
//---------------------------------------------------------------------------
void MapListMgmt::AddToUnfilter(TListBox *FilteredList)
{
   int numcount;
   int DecimalNumber, i;
   NUM_Mgmt_Iterator_Type iter;

   // get the number of items in the TListBox
   numcount = FilteredList->Items->Count;
   // loop through each item

   for (i=0; i < numcount; i++) {
      // if the item is selected then process it

      if (FilteredList->Selected[i]) {
         // change the string to a decimal integer
         DecimalNumber = ChangeToInteger(FilteredList->Items->Strings[i].c_str());
         // see if the number is in the filtered list
         iter = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].find(DecimalNumber);

         // if the number was found in the filtered list
         if (iter != NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].end()) {
            // copy the item to the unfiltered list
            NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].insert(NUM_Mgmt_Entry(iter->first,
                                                                    iter->second));
            // delete the item from the filtered list
            NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].erase(iter);
         }

         // delete this item from the TListBox
         FilteredList->Items->Delete(i);
         // decrement the number of items because we deleted one
         numcount--;
         // decrement the loop variable so we do not skip over an item
         i--;
      }  // end if (FilteredList->Selected[i])

   }  // end of for loop

}

//---------------------------------------------------------------------------
// Function FilterStatus is called to find out whether the number is in
// the filtered list, unfiltered list, or no list at all.  The function returns
// the result to the calling function.
//---------------------------------------------------------------------------
int MapListMgmt::FilterStatus(int ThisNumber)
{
   // returns whether a number is filtered, unfiltered or not in the list
   int filterstatus = NUM_MGMT_NOT_IN_LIST;

   // if the unfiltered list is not empty
   if (!NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].empty()) {

      // search for the number in the unfiltered list
      if (NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].find(ThisNumber) !=
          NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].end()) {
         // if it was found indicate it
         filterstatus = NUM_MGMT_UNFILTERED;
      }

   }

   // if the number was not found and the filtered list is not empty
   if ((filterstatus == NUM_MGMT_NOT_IN_LIST) &&
       (!NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].empty())) {

      // search for the number in the filtered list
      if (NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].find(ThisNumber) !=
          NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].end()) {

         // if it was found indicate it
         filterstatus = NUM_MGMT_FILTERED;
      }

   }

   return (filterstatus);
}

//---------------------------------------------------------------------------
// Function UNfilterAll copies all the numbers into the unfiltered list.
// The function goes through the filtered list and copies each item to the
// unfiltered list.
//---------------------------------------------------------------------------
void MapListMgmt::UnfilterAll()
{
   // Unfilters all filtered numbers
   NUM_Mgmt_Iterator_Type iter, iterend, tempiter;

   // if there is something in the filtered list
   if (!NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].empty()) {

      // set the iterators to the first and last items
      iter = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].begin();
      iterend = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].end();

      while (iter != iterend) {

         // copy the item from one list to the other
         NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].insert(NUM_Mgmt_Entry(iter->first,
                                                            iter->second));
         tempiter = iter;
         iter++;
         // delete the item from the list it came from
         NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].erase(tempiter);
      }

   }

}

//---------------------------------------------------------------------------
// Function FilterAll copies all the numbers into the filtered list.  The
// function goes through the unfiltered list and copies each item to the
// filtered list.
//---------------------------------------------------------------------------
void MapListMgmt::FilterAll()
{
   // Filters all unfiltered numbers
   NUM_Mgmt_Iterator_Type iter, iterend, tempiter;

   // if there is something in the unfiltered list
   if (!NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].empty()) {

      // set the iterators to the first and last items
      iter = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].begin();
      iterend = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].end();

      while (iter != iterend) {

         // copy the item from one list to the other
         NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].insert(NUM_Mgmt_Entry(iter->first,
                                                            iter->second));
         tempiter = iter;
         iter++;
         // delete the item from the list it came from
         NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].erase(tempiter);
      }

   }

}

//---------------------------------------------------------------------------
// Function GetListSize is called to obtain the number of items in a list.  The
// input paramter Use_List indicates which list (filtered or unfiltered) to
// use.  The number of items in the list is returned to the calling function.
//---------------------------------------------------------------------------
int MapListMgmt::GetListSize(int Use_List)
{
   NUM_Mgmt_Iterator_Type iter, iterend;
   int i, numitems;

   numitems = 0;

   // set the iterators to the first and last items
   if (Use_List == NUM_MGMT_FILTERED_LIST) {

      if (!NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].empty()) {
         numitems = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].size();
      }

   }
   else {

      if (!NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].empty()) {
         numitems = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].size();
      }

   }

   return (numitems);
}

//---------------------------------------------------------------------------
// Function GetListItem gets the next item from the specified list.  The input
// parameter Use_List indicates which list to use.  The input parameter reset
// indicates whether to start at the first item or to continue from the last
// item returned.  This function returns a boolean to indicate a number
// is being returned.  If the return value is true then the value in ItemValue
// is valid.
//---------------------------------------------------------------------------
bool MapListMgmt::GetListItem(int *ItemValue, int Use_List, bool reset)
{
   int i, numitems;
   bool returnstatus = false;

   // set the iterators to the first and last items
   if (Use_List == NUM_MGMT_FILTERED_LIST) {

      if (!NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].empty()) {

         if (reset) {
            FilterIter = NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].begin();
         }

         if (FilterIter != NUM_Mgmt_List[NUM_MGMT_FILTERED_LIST].end()) {
            *ItemValue = FilterIter->first;
            FilterIter++;
            returnstatus = true;
         }

      }

   }
   else {

      if (!NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].empty()) {

         if (reset) {
            UnFilterIter = NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].begin();
         }

         if (UnFilterIter != NUM_Mgmt_List[NUM_MGMT_UNFILTERED_LIST].end()) {
            *ItemValue = UnFilterIter->first;
            UnFilterIter++;
            returnstatus = true;
         }

      }

   }

   return (returnstatus);
}
//---------------------------------------------------------------------------


