unit ManViewer;

{ ************************************************************************** }
{                                                                            }
{ This unit implements a man page viewer for use in Linux-based CLX          }
{ applications, including Kylix itself. It has not been extensively          }
{ tested on other forms of unix, or on forms of Linux except for RedHat,     }
{ and there are no guarantees that the assumptions about file locations, etc }
{ contained within this file are correct. When in doubt, *always* set the    }
{ environment variables rather than relying on the default values; the       }
{ environment variables are guaranteed to be checked first.                  }
{                                                                            }
{ This unit is provided as an example, and you are free to modify and        }
{ redistribute it as desired provided that the copyright notice remains.     }
{                                                                            }
{ Copyright (c) 2000, 2002 Borland Software Corporation                      }
{                                                                            }
{ ************************************************************************** }

interface

{ This file, in fact, has no interface. }

{ ========================================================================== }

implementation

uses HelpIntfs, Classes, SysUtils, LibC;

type
  { TManPageViewer.
    TManPageViewer implements the ICustomHelpViewer interface, the basic
    interface used to recieve requests from the Help Manager. It does not
    implement any of the extended interfaces. }

  TManPageViewer = class(TInterfacedObject, ICustomHelpViewer)
  private
    FHelpStrings : TStringList;
    FLastQuery : String;
    FViewerID : Integer;
    ChildPid : Integer;
    procedure ProcessHelpStrings(StringBuf: PChar; HelpString: String);
    procedure KillChild;
  public
    FHelpManager : IHelpManager;
    constructor Create;
    procedure InternalShutDown;
    { ICustomHelpViewer }
    function GetViewerName : String;
    function UnderstandsKeyword(const HelpString: String): Integer;
    function GetHelpStrings(const HelpString: String): TStringList;
    function CanShowTableOfContents : Boolean;
    procedure ShowHelp(const HelpString: String);
    procedure ShowTableOfContents;
    procedure NotifyID(const ViewerID: Integer);
    procedure SoftShutDown;
    procedure ShutDown;
    property HelpManager : IHelpManager read FHelpManager write FHelpManager;
    property ViewerID : Integer read FViewerID;
    destructor Destroy; override;
  end;

{ global instance of the ManPageViewer which can be handed off to the
  Help Manager at registration time. }
var
  HelpViewer : TManPageViewer;

const
  { man and its switches }
  ExeName = '/usr/bin/man';
  AllSwitch = '-a'; { show all man pages on a topic }
  WhereSwitch = '-w'; { where is the man page located? }

  { xterm and its switches }
  ViewerName = 'xterm'; // 'xterm'; 
  MoreBugSwitch = '-cu'; { work around pager bug }
  ExecSwitch = '-e'; { launch xterm running a program. 
                       must be the last argument to xterm;
                       everything which follows it is an
                       argument to the exec call. }
  TitleSwitch = '-T'; {set the title of the window }
  ViewerTitle = 'Kylix man page viewer';

  { signal used to kill child processes }
  KillSignal = SIGINT;

  { error conditions}
  sFatalFork = 'Unable to fork(). Please consult the disaster manual.';
  sNoTableOfContents = 'Unable to provide table of contents for man pages.';

{ -------------------------------------------------------------------------- }
{ TManPageViewer }

constructor TManPageViewer.Create;
begin
  inherited Create;
end;

{ ProcessHelpStrings is used internally by the man viewer to decompose
  information returned by a query. The Man Viewer invokes man with
  'man -w foo'; man returns something along the lines of
  /usr/man/man#/foo.#.gz. The interesting thing here is the #, which
  indicates what section of the manual the page is in. 

   The overall operation of the function is to morph 
   '/usr/man/man#/foo.#.gz' into 'foo (#) (ViewerName)' to pass back
   to the Help Manager in cases where more than one match is found,
   enabling the Help Manager to ask the user of the application to
   decide which match matters. }

procedure TManPageViewer.ProcessHelpStrings(StringBuf: PChar; 
                                            HelpString: String);
var
  bufptr, lineptr, valptr, delim: PChar;
  searching: boolean;
  addstr : String;
begin
  { copy the pointer so that you still have a pointer to the beginning;
    strsep modifies the pointer }
  bufptr := StringBuf;

  { the passed-in string should consist of some number of lines
    of the format: /filepath/HelpString.section.gz, plus a trailing 
    empty line or two. }

  searching := true;
  while searching do
  begin
    { search for an end-of-line in the passed string. }
    delim := #10#13;
    lineptr := strsep(@bufptr, delim);
    if (lineptr = nil) then
    begin
     { if an end-of-line wasn't found, then the string is somehow broken.
       this may contain a bug --- if the returned data was more than 1024
       bytes, then we may be breaking in the middle of a line and losing
       help strings. I can't find any cases that demonstrate it, tho. }
     searching := false;
    end else
    begin
     delim := '.';
     { call strsep twice to retrieve the data between the first
       and second dots. }
     strsep(@lineptr, delim);
     valptr := strsep(@lineptr, delim);
     { if there is nothing after the second dot, then something strange
       happened. }
     if valptr <> nil then
     begin
       { build a string that looks sorta like: 'foo (section #) (man page)' }
       addstr := HelpString + ' (' + valptr + ') (' + GetViewerName + ')';
       { and add it to a string list which can get passed back to the
         Help Manager. }
       FHelpStrings.Add(addstr);
     end;
    end;
  end;
end;

{ KillChild is used internally to kill externally spawned man viewers.
  This is called whenever a new viewer is about to be spawned, on
  shutdown, and at certain times when the Help Manager demands that
  it be done. }

procedure TManPageViewer.KillChild;
begin
  if ChildPid <> 0 then
  begin
     kill(ChildPid, KillSignal);
     { the following is to clean up zombie processes. note that this
       can theoretically cause problems if the process has already
       been cleaned up by someone else, and the pid was reassigned. }
     waitpid(ChildPid, nil, WNOHANG or WUNTRACED);
     ChildPid := 0;
  end;
end;

{ InternalShutDown is called when the unit is exiting and the Help Manager
  needs to be informed. }
  
procedure TManPageViewer.InternalShutDown;
begin
  KillChild;
  if Assigned(FHelpManager) then FHelpManager.Release(ViewerID);
  ShutDown;
end;

{ --------------------------------------------------------------------------- }
{ TManPageViewer --- ICustomHelpViewer }

{ GetViewerName is called by the Help Manager to provide a unique identifier
  which users can use to select between multiple help systems. }

function TManPageViewer.GetViewerName;
begin
  Result := ExeName;
end;

{ UnderstandsKeyword is called by the Help Manager to ask if a particular
  keyword is supported by a particular Help Viewer.

  The Man Viewer responds to this question by asking man, with a query
  of 'man -w [keyword]'. }

function TManPageViewer.UnderstandsKeyword(const HelpString: String): Integer;
var
  { variables used when execing }
  SuccDescr, ErrDescr : TPipeDescriptors;
  pid: Integer;
  Args : array of PChar;
  DescriptorSet : TFDSet;
  WaitTime : TTimeVal;
  { variables used to read from the socket }
  PipeStream : THandleStream;
  ReadBuf : Pointer;
  BytesRead: Integer;
  Reading : Boolean;
begin
  Result := 0;
  { free any leftover help strings and set up important boolean}
  if FHelpStrings <> nil then FHelpStrings := nil;

  { man will tell us about help if we ask it. nicely, with the -a -w switches.
    the thing to do is to redirect man's output to two pipes --- one for stdout,
    one for stderr. the identity of the pipe which gets invoked will tell us
    if help was found --- yes on stdout, no on stderr. }
  
  { set up arguments to pass to exec }
  SetLength(Args, 5);
  Args[0] := ExeName;
  Args[1] := AllSwitch;
  Args[2] := WhereSwitch;
  Args[3] := PChar(HelpString);
  Args[4] := nil;

  { set up two pipes to communicate with. }
  pipe(SuccDescr);
  pipe(ErrDescr);

  pid := fork;
  if pid = 0 then
  begin
    { close the read descriptors, because we don't care. }
    __close(SuccDescr.ReadDes);
    __close(ErrDescr.ReadDes);
    { dup the write descriptors onto stdout and stderr }
    dup2(SuccDescr.WriteDes, stdout);
    dup2(ErrDescr.WriteDes, stderr);
    __close(stdin);
    { exec man and ask it if there's help for this topic }
    execvp(PChar(Args[0]), @Args[0]);
    _exit(1);
  end
  else begin
    if pid = -1 then
    begin
      { this probably means that something has gone very, very, wrong. }
      raise EHelpSystemException.Create(sFatalFork);
    end else
    begin
      { set up information about file descriptors }
      { set up a blocking select that will time out at a certain
        point --- say, 5 seconds. this should be long enough for
        man to operate and then release the selector. }
      WaitTime.tv_sec := 10;
      WaitTime.tv_usec := 0;

      { cast the descriptors to TSocket, because the pascal declaration
        of the function demands that --- but note that since TSocket is
        an integer, and the file descriptor is an integer, this is OK. }
      FD_ZERO(DescriptorSet);
      FD_SET(TSocket(SuccDescr.ReadDes), DescriptorSet);
      FD_SET(TSocket(ErrDescr.ReadDes), DescriptorSet);

      { check and see which file descriptor has been written to. }
      select(__FD_SETSIZE, @DescriptorSet, nil, nil, @WaitTime);

      { if STDERR was written to, no help was found, and we exit.
        if STDOUT was written to, then help was found, and we're ok. :)}
      if FD_ISSET(TSocket(SuccDescr.ReadDes), DescriptorSet) then
      begin
        { yay! we found help. :) }
        { now, set up the help strings for future use }
        if FHelpStrings = nil then FHelpStrings := TStringList.Create;
        { read the pipe and decompose man's output into something we want. }
        PipeStream := THandleStream.Create(SuccDescr.ReadDes);
        ReadBuf := Libc.malloc(1024);
        memset(ReadBuf, 0, 1024);
        Reading := true;
        while Reading do
        begin
          BytesRead := PipeStream.Read(ReadBuf^, 1024);
          if (BytesRead < 1024) then Reading := false;
          ProcessHelpStrings(ReadBuf, HelpString);
          memset(ReadBuf, 0, 1024);
        end;
        Libc.free(ReadBuf);
        PipeStream.Free;
        Result := FHelpStrings.Count;
        FLastQuery := HelpString;
      end else if not (FD_ISSET(TSocket(ErrDescr.ReadDes), DescriptorSet)) then
      begin
        { danger will robinson! for some reason the read descriptor didn't
          reap any information. oops. go ahead and create an empty string
          list so the calling function isn't confused, and output
          a warning string. }
        Writeln('Warning: man invocation failed.');
        if FHelpStrings = nil then FHelpStrings := TStringList.Create;
      end;

      { make sure man exited correctly. }
      waitpid(pid, nil, WNOHANG or WUNTRACED);
    end;
  end;
  { close all the descriptors }
  __close(SuccDescr.WriteDes);
  __close(ErrDescr.WriteDes);
  __close(SuccDescr.ReadDes);
  __close(ErrDescr.ReadDes);
end;

{ GetHelpStrings is called by the Help Manager when more than one Viewer
  supports help for a particular keyword, and it needs to display a UI
  element to allow the user to select a keyword. The string list for
  this got built during UnderstandsKeyword, and it is guaranteed that
  this will not be called except *after* Understands Keyword, so just
  return the internal string list ...}

function TManPageViewer.GetHelpStrings(const HelpString: String): TStringList;
begin
  Result := FHelpStrings;
end;

{ CanShowTableOfContents is called by the Help Manager when it wants Viewers
  to display Tables of Contents. While it's not strictly true that man can't
  do this --- the Viewer could walk through the directories in the man path
  and list the files --- it's a lot of work for little gain, so this Viewer
  always refuses. }

function TManPageViewer.CanShowTableOfContents: Boolean;
begin
  Result := false;
end;

{ ShowTableOfContents is called by the Help Manager if a Viewer says it can
  show a table of contents. Since the Man Viewer always refuses, this 
  function should *never* be called, and so it's OK to just throw. }
procedure TManPageViewer.ShowTableOfContents;
begin
  raise EHelpSystemException.Create(sNoTableOfContents);
end;

{ ShowHelp is called by the Help Manager if a Viewer says it has help
  for a particular keyword, and either (a) only it can provide help;
  (b) the user has selected it to show the help; or (c) it was the
  first registered viewer  (if the user didn't select anyone. }

procedure TManPageViewer.ShowHelp(const HelpString: String);
var
  { variables used to understand HelpString }
  KeywordEnd, Section, CompResult, CompString, Comparator: PChar;
  { variables used when execing }
  Args : array of PChar;
  pid : Integer;
  retval: Integer;
begin 
  { if there's still a help window open from the last time, kill it. }
  KillChild;

  { set up arguments }
  SetLength(Args, 9);
  Args[0] := ViewerName; 
  Args[1] := MoreBugSwitch;
  Args[2] := TitleSwitch;
  Args[3] := ViewerTitle;
  Args[4] := ExecSwitch;
  Args[5] := ExeName;
  Args[6] := AllSwitch;
  Args[7] := PChar(HelpString);
  Args[8] := nil;

  { the invoker may have passed in a string that came from ProcessStrings.
    if so, then decompose it here, and change the arguments as appropriate. }
  CompString := PChar(HelpString);
  Comparator := Libc.malloc(2);
  Comparator[0] := '(';
  Comparator[1] := #0;
  CompResult := strstr(CompString, Comparator);
  Libc.free(Comparator);

  { note that there's a bug in that --- if the man string contained a '('
    naturally, then this Viewer will assume the string needs decomposition. }

  if (CompResult <> nil) then
  begin
   { instead of -w, use the correct number }
   Section := Libc.malloc(2);
   KeywordEnd := AnsiStrPos(PChar(HelpString), '(');
   { StrLCopy requires null termination; if we're going to manipulate
     things to get that, we might as well do direct char-per-char
     copying. }
   Section[0] := KeywordEnd[1];
   Section[1] := #0;
   Args[6] := Section;
   { #DEFINE DUMB_HACK_BY_TIRED_PROGRAMMER }
   Args[7] := PChar(FLastQuery);
  end
  else begin
   Section := nil;
  end;

  pid := fork;
  if pid = 0 then
  begin
    { invoke xterm in such a fashion that it runs man on the page
      we're interested in. }
    retval := execvp(PChar(Args[0]), @Args[0]);
    _exit(1);
  end 
  else begin
    if pid = -1 then
    begin
      raise EHelpSystemException.Create(sFatalFork);
    end 
    else begin
      { save the pid for later killing }
      ChildPid := pid;
    end;
  end;

  if Section <> nil then Libc.free(Section);
end;

{ NotifyID is called by the Help Manager after successful registration
  to provide the Viewer with a unique cookie which identifies it in
  communications with the Help Manager. }

procedure TManPageViewer.NotifyID(const ViewerID: Integer);
begin
  FViewerID := ViewerID;
end;

{ SoftShutDown is called by the Help Manager when someone asks it to
  shut down all external manifestations of the Help System without
  actually shutting the Help System down. }

procedure TManPageViewer.SoftShutDown;
begin
  KillChild;
end;

{ ShutDown is called when the Help Manager issues a shut down order. It
  does not need to be informed, as it originated the command. }

procedure TManPageViewer.ShutDown;
begin
  KillChild;
  if Assigned(FHelpManager) then FHelpManager := nil;
end;

destructor TManPageViewer.Destroy;
begin
  inherited Destroy;
end;

{ ========================================================================== }

initialization
  if not Assigned(HelpViewer) then 
  begin
    HelpViewer := TManPageViewer.Create;
    HelpIntfs.RegisterViewer(HelpViewer, HelpViewer.FHelpManager);
  end;
finalization
  if Assigned(HelpViewer) then
  begin
    HelpViewer.InternalShutDown;
  end;
end.
