/*
 *  History: Original by Allen Bauer
 *
 *   Modified for install tests by Dmitri Smirnov
 *   Modified for redistribution by Darren Kosinski
 *
 *   (c) Copyright 2000 Borland
 *
 *   Arguments:
 *    The program may accept one optional argument -q which puts it into
 *    silent mode. Thus there will be no output on console and it will not solicit output
 *    from the user.
 *    Silent functionality was requested by the installer. In the silent mode the program
 *    only checks for the loader design issue but does not perform other checks because the installer
 *    performs them anyway
 *
 *  Checks:
 *     glibc loader bug test
 *     Kernel version must be >= 2.2
 *     glibc version must be >= 2.1.2
 *     /usr/lib/libjpeg.so.62 >= 6.2.0
 *
 *
 *    Program exit codes:
 *       0 - Success
 *       1 - Loader flaw
 *       2 - Kernel verion is low
 *       3 - glibc version is low
 *       4 - libjpeg not found or version is low
 *       5 - Invalid cmd argument
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>
#include <link.h>
#include <dlfcn.h>
#include <stdarg.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <gnu/libc-version.h>

/* Supplementary libraries */
#define LIBA_LIB  "lib-a.so"
#define LIBB_LIB  "lib-b.so"
#define LIBC_LIB  "lib-c.so"

/* System libraries locations */
#define LIBC_SYSPATH  "/lib/"
#define LIBC_SYSNAME  "libc.so.6"
#define LIBJPEG_SYSPATH "/usr/lib/"
#define LIBJPEG_SYSNAME "libjpeg.so.62"

/* Constants */
#define LIBC_MAJOR 2
#define LIBC_MINOR 1
#define LIBC_BUILD 2

/* Program exit codes */
#define EX_LOADER_FLAW		1
#define EX_KERNEL_VERSION	2
#define EX_LIBC_VERSION		3
#define EX_JPEG_VERSION		4
#define EX_USAGE_ERROR		5

#define PATH_DELIM		'/'

#define MAX_PATH_LEN		256


/* Used for func addresses */
typedef void (*dl_function)(int);

/* Flag silent mode */
int    bSilent = 0;

/* flag for verbose */
int verbose = 0;

/* =============== Protos ============================================= */

/* Parse arguments */
int parse_args(int, char**);

/* Report OS error on the screen */
int report_error(const char* mes);

/* Display objects in the memory */
void dump_shared_objects(void);

/* Wrapper to privide a silent output */
int  log_message(FILE*, const char*, ...);

/* Wrapper for silent mode 
void  get_any_key(void); */

/* Check if the library is currently loaded */
int  IsLoaded(const char* libname);

/* Checks kernel version against requested */
int check_kernel_version(int major, int minor);

/* The fucntion checks shared object version using SO name */
int check_so_version(const char* libpath, const char* libname, int major, int minor, int revision);

/* The fucntion accepts a pointer to the first version number and returns
   version numbers
   the return value of the function is a number of version numbers converted
  */
int get_so_version(const char* verstr, int* major, int* minor, int* revision);

/* Function to check if the version of libc is adequate */
int check_libc(void);

/* This function will test dynamic loading flaw */

int test_dynamic(void);

/* ===================  Implementation ================================== */

int parse_args(int argc, char** argv)
{
  int c;


  if(argc == 1)
   return 0;

  while ((c=getopt(argc, argv, "qv")) != EOF) {
    switch (c) {
    case 'q':
      bSilent = 1;
      break;

    case 'v':
      if (! bSilent) {
	verbose = 1;
      }
      break;

    default:
      log_message(stderr, "BORPRETEST: Only know -q or -v\n");
      return EX_USAGE_ERROR;
    }
  }
  return 0;
}

/* Permission to continue 
void  get_any_key(void)
{
  if(bSilent)
   return;
  log_message(stdout, "Press any key...\n");
  getchar();
} */

int report_error(const char* pref)
{
/* Print out OS error */
  char buf[1024];
  if(bSilent)
    return 0;

  strcpy(buf, pref);
  strcat(buf, strerror(errno));
  strcat(buf, "\n");
  return log_message(stderr, buf);
}

int log_message(FILE* fout, const char* format, ...)
{
  int out;
  va_list ap;

  if(bSilent)
    return 0;

  va_start(ap, format);
  out = vfprintf(fout, format, ap);
  va_end(ap);
  return out;
}

int IsLoaded(const char* libname)
{
   struct link_map *lm;
   char            *p;
   for (lm = _r_debug.r_map; lm; lm = lm->l_next)
   {
       p = strrchr(lm->l_name, PATH_DELIM);
       if(p != NULL)
         p++;
       else
         p = lm->l_name;

       if(strcmp(libname, p) == 0)
        return 1;
  }
   return 0;
}

int check_kernel_version(int major, int minor)
{
  /*
    The function ensures that the kernel version is
    equal or greater than requested
  */
  struct utsname uname_info;
  int   maj_this, min_this;
  char  *p, *pb;

  if(uname(& uname_info) == -1)
  {
     report_error("Uname: ");
     return 0;
  }
#ifdef _DEBUG
  fprintf(stderr, "Kernel version is %s\n", uname_info.release);
#endif

  pb = uname_info.release;
  p = strchr(uname_info.release, '.');
  if(p == NULL)
    return -1;

  *p = '\0';
   maj_this = atoi(pb);
   if(maj_this < major)
    return 0;

   if(maj_this > major)
    return 1;

   pb = p + 1;
   p = strchr(pb, '.');
   if(p == NULL)
     return -1;

   *p = '\0';
   min_this = atoi(pb);
   if(min_this < minor)
    return 0;

   return 1;
}

int check_so_version(const char* libpath, const char* libname, int major, int minor, int revision)
{
 /* Function checks library for existance. Checks the version.
    if it is link then it follows the link
  */
  char   fullpath[MAX_PATH_LEN];
  struct stat libinfo;
  char *p;
  int   maj_this, min_this, rev_this;
  char  sep;

  maj_this = min_this = rev_this = -1;

  strcpy(fullpath, libpath);
  strcat(fullpath, libname);

  if (verbose)
    log_message(stdout, "Checking %s\n", fullpath);

  if(lstat(fullpath, & libinfo) == -1)
  {
    report_error("Stat: ");
    return 0;
  }

  /* Follow the link */
  if(S_ISLNK(libinfo.st_mode))
  {
     /* Get the link path */
     char  linkname[MAX_PATH_LEN];
     int n;

     n = readlink(fullpath, linkname, MAX_PATH_LEN);
     if(n == -1)
     {
        report_error("readlink: ");
        return 0;
     }
     linkname[n] = '\0';
     if (verbose)
       log_message(stdout, "%s resolves to %s\n", fullpath, linkname);
     return check_so_version(libpath, linkname, major, minor, revision);
  }
  /* Find out where the version string starts */
  sep = '-';
  p = (char*)libname;
  while(1)
  {
     p = strchr(p, sep);
     if(p == NULL)
     {
       if(sep == '.')
         return 0;
       sep = '.';
       p = (char*)libname;
       continue;
     }
     p++;
     if(isdigit(*p))
       break;
  }

   if(get_so_version(p, & maj_this, & min_this, & rev_this) < 3)
     return 0;
#ifdef _DEBUG
   fprintf(stderr, "%s version is %d.%d.%d\n", libname, maj_this, min_this, rev_this);
#endif

   if(maj_this < major)
     return 0;

   if(maj_this > major)
     return 1;

   if(min_this < minor)
     return 0;

   if(min_this > minor)
     return 1;

   if(rev_this < revision)
      return 0;

   return 1;
}

int get_so_version(const char* verstr, int* major, int* minor, int* revision)
{
   char num[10];
   int  i, n;

   n = 0;
   while(n < 3)
   {
     num[0] = '\0';
     for(i = 0; i < 10 && *verstr != '\0'; i++)
     {
       if(*verstr == '.')
       {
          verstr++;
          break;
       }
       if(isdigit(*verstr))
        num[i] = *verstr;

       verstr++;
     }
     num[i] = '\0';
     if(i == 0) /* Nothing is read */
      break;
     switch(n)
     {
       case 0:
	 if (verbose)
	   log_message(stdout, "Major version is %s\n", num);
          *major = atoi(num);
           break;
       case 1:
	 if (verbose)
	   log_message(stdout, "Minor version is %s\n", num);
          *minor = atoi(num);
           break;
       case 2:
	 if (verbose)
	   log_message(stdout, "Revision version is %s\n", num);
          *revision = atoi(num);
           break;
       default:
           assert(0 != 0);
     }
     n++;
   }
  return n;
}

int check_libc()
{
  int major = LIBC_MAJOR;
  int minor = LIBC_MINOR;
  int rev = LIBC_BUILD;

  if (get_so_version(gnu_get_libc_version(), &major, &minor, &rev) >= 2) {
    if (major > LIBC_MAJOR) {
      return 1;
    } else if (major == LIBC_MAJOR && minor > LIBC_MINOR) {
      return 1;
    } else if (major == LIBC_MAJOR && minor == LIBC_MINOR && rev >= LIBC_BUILD) {
      return 1;
    }
  }
  return 0;
}

void dump_shared_objects(void)
{
   /* Display shared objects loaded */
   struct link_map *lm;
   if (bSilent || ! verbose)
     return;
   log_message(stdout, "   Name                                     Address     Handle\n");
   log_message(stdout," -------------------------------------------------------------------\n");
   for (lm = _r_debug.r_map; lm; lm = lm->l_next)
   {
   	log_message(stdout," %-40s  0x%.8x  0x%.8x\n", lm->l_name, (int)lm->l_addr, (int)lm);
   }
}

int test_dynamic(void)
{
 /* Check the system whether it has a dynamic loading flaw
    Returns 0 for FAIL 1 for PASS
  */
   void *a_handle;
   void *b_handle;
   dl_function a_function;
   dl_function b_function;
   char *err;
   int  result;
   
   if (verbose) {
     log_message(stdout,"This will test whether libc installed on your system has\n");
     log_message(stdout,"a design flaw of the dynamic loading of\n");
     log_message(stdout,"shared objects.  By dynamicly loading shared object 'a' that depends\n");
     log_message(stdout,"on shared objects 'b' and 'c', then dynamically loading shared object 'b'\n");
     log_message(stdout,"which depends on 'c', shared object 'c' will remain in memory when 'a' \n");
     log_message(stdout,"and 'b' are closed.\n");
     log_message(stdout,"\nThis is what is in memory now:\n");
     dump_shared_objects();
     log_message(stdout,"Loading shared object lib-a.so\n");
    }
    a_handle = dlopen("lib-a.so", RTLD_LAZY);
    if (!a_handle)
    {
    	log_message(stdout, "%s\n", dlerror());
    	exit(1);
    }    	
    if (verbose) {
      log_message(stdout,"And this is what is now in memory\n");
      dump_shared_objects();
      log_message(stdout,"Getting address of 'a_function' and calling it\n");
    }
    (void*)a_function = dlsym(a_handle, "a_function");
    err = dlerror();
    if (err)
    {
    	log_message(stdout,"%s\n", err);
    	exit(1);
    }
    a_function(bSilent);
    if (verbose)
      log_message(stdout,"Now loading shared object lib-b.so\n");
    b_handle = dlopen("lib-b.so", RTLD_LAZY);
    if (!b_handle)
    {
 		log_message(stdout,"%s\n", dlerror());
 		exit(1);
    }
    if (verbose) {
      log_message(stdout,"Again, this is what is in memory\n");
      dump_shared_objects();
      log_message(stdout,"Getting address of 'b_function' and calling it\n");
    }
    (void*)b_function = dlsym(b_handle, "b_function");
    err = dlerror();
    if (err)
    {
		log_message(stdout,"%s\n", err);
		exit(1);
    }
    b_function(bSilent);
    if (verbose) {
      log_message(stdout,"Now this is where things begin to break down.  The lib-b.so handle\n");
      log_message(stdout,"will be closed first, then the lib-a.so handle will be closed.  Expected\n");
      log_message(stdout,"behaviour should be that the objects in memory return to the state prior\n");
      log_message(stdout,"loading lib-a.so\n");
      log_message(stdout,"Closing b_handle\n");
    }
    dlclose(b_handle);
    dump_shared_objects();
    if (verbose) {
      log_message(stdout,"So far so good\n");
      log_message(stdout,"Closing a_handle\n");
    }
    dlclose(a_handle);
    dump_shared_objects();
    result = IsLoaded(LIBC_LIB);
    if(result)
    {
      if (verbose) {
	log_message(stdout,"WHOOPS!!!  Lookie there!  lib-c.so is still hanging around! This is bad...\n");
	log_message(stdout,"Thanks for your time\n");
      }
      return 0;
    }
    return 1; /* Return PASS */
}


int main(int argc, char **argv)
{
    int result;
    int fail = 0;

    result = parse_args(argc, argv);

   log_message(stdout, "Borland Kylix System Compatibility Test\n\n");
   if(result)
    return result;

   result = EXIT_SUCCESS;
   log_message(stdout, "Checking loader....");
   if(test_dynamic() == 0) {
     log_message(stdout, "FAILED\n");
     result = EX_LOADER_FLAW;
     if (bSilent)
       return (result);
     fail = 1;
     } else
       log_message(stdout, "OK\n");

   if(bSilent)
     return EXIT_SUCCESS;
   
   log_message(stdout, "Checking kernel >= 2.2....");
   if(check_kernel_version(2, 2) == 0)
   {
     log_message(stdout, "FAILED\n");
     if (! fail) {
       result = EX_KERNEL_VERSION;
       fail = 1;
     }
   } else
     log_message(stdout, "OK\n");

   log_message(stdout, "Checking libc >= 2.1.2....");
   if (check_libc()) {
     log_message(stdout, "OK\n");
   } else {
     log_message(stdout, "FAILED\n");
     if (! fail)
       fail = 1;
   }

   log_message(stdout, "Checking libjpeg >= 6.2.0....");
   if(check_so_version(LIBJPEG_SYSPATH, LIBJPEG_SYSNAME, 62, 0, 0) == 0)
   {
      log_message(stdout, "FAILED\n");
      if (! fail) {
	result = EX_LIBC_VERSION;
	fail = 1;
      }
   } else
     log_message(stdout, "OK\n");

   if (! fail) {
     log_message(stdout, "\nLooks GOOD !!!\nThis system should be able to run Borland Kylix!\n\n");
   } else {
     log_message(stdout, "\nThis system is not compatible with Borland Kylix. Please see the\n");
     log_message(stdout, "documents in this directory for information on how to upgrade your\n");
     log_message(stdout, "system.\n\n");
   }

   return (result);
}



























