/* **************************************************************************
 *                                                                          *
 * diag.c --- Run-time support for debug facilities.                        *
 *                                                                          *
 * Copyright (c) 1996, 2001, 2002 Borland Software Corporation                    *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#define COMPILING_DIAG
#include "diag.h"

/*
	The initializations below look funny, don't they?  They're there to satisfy
	a limitation in the Delphi object file importer.
*/
static FILE *_dbfp = (FILE *)1;

/*
	The number of times debug_init() since the last debug_shutdown().
	If this value is greater than one, check the logs for a hint.
*/
static int _dbinits;

static char *current_class = 0;

typedef enum { false, true } bool;

/*
 * Function:  glob_match
 *
 * Purpose:	
 *	 See if one string matches another, allowing for globbing
 *	 characters.
 *
 * Returns:	
 *	 'true' if the two strings match, 'false' if not.
 *
 * Notes:
 *   ilink32.*list.info != ilink32.filelist.info because it tries to
 *   match against ilink32.*lelist.info (the first occurance of 'l' on
 *   the line).  The matching strategy needs to continue searching
 *   until the end of the line, more like a state machine would.
 */
static bool glob_match(const char *glob, const char *str)
{
	const char *p = glob, *q = str;
	for	(; *p && *q; p++, q++)
	{
		if	(*p == '*')
		{
			char next = *(p + 1);
			if	(! next)
				return true;
			for	(; *q && *q != next; q++) ;
			if	(! *q)
				return false;
			p++;
		}
		else if	(*p != '?' && *p != *q)
			return false;
	}
	return (! *q && *p == '*' && ! *(p + 1)) || *p == *q;
}


/*
 * Function:  debug_class_enabled
 *
 * Purpose:	
 *	 Test whether a certain debug class has been enabled.  To enable a
 *	 class, set the environment variable DEBUG_CLASS to a series of a
 *	 semi-colon separated globbing patterns that will match to a
 *	 class.
 *
 * Returns:	
 *	 1 if the class is currently enabled, 0 if not.
 */
int debug_class_enabled(char *dbcls)
{
	static enum { NOT_READ, ENABLED, DISABLED } state = NOT_READ;
	static char *enabled_classes[MAX_CLASSES];
	static int   class_count 			  = 0;
	int	     index;
	char	    *p;
		
	if	(state == NOT_READ)
	{
		char *env_ptr = getenv("DEBUG_CLASS");
		state = env_ptr ? ENABLED : DISABLED;
		if	(env_ptr)
		{
			for	(p = strtok(env_ptr, ";:"); p; p = strtok(NULL, ";:"))
			{
				enabled_classes[class_count] = strdup(p);
				class_count++;
			}
		}
	}
	else if	(state == DISABLED)
		return 0;
		
	for	(index = 0; index < class_count; index++)
	{
		if	(glob_match(enabled_classes[index], dbcls))
		{
			current_class = dbcls;
			return 1;
		}
	}
	return 0;
}

static	FILE *	getSTDOUT(void)
{
#ifdef	__linux__
	/* With GNU libc, stdout is a variable with external linkage.  We
	   cannot refer to it in Pascal, because we can't declare external
	   variables.  */
	#undef stdout
	#define stdout do_not_use_stdout_see_function_getSTDOUT

	/* This doesn't work well:
		return fdopen (1, "w");
	   if someone closes the file, we also end up losing the "real"
	   standard output file descriptor as well. */
	int fd = dup(1);
	return fd >= 0 ? fdopen(fd, "w") : fopen("/dev/tty", "w");
#else
	return stdout;
#endif
}

/*
 * Function:  debug_file
 *
 * Purpose:	
 *	 Allow external modules to redirect to the same file
 *   the debugging output stream.
 *
 * Returns:	
 *	 file pointer
 */
FILE *debug_file(void)
{
  return _dbfp;
}

/*
 * Function:  debug_init
 *
 * Purpose:	
 *	 Initialize the debugging system.  Primarily this means setting
 *   the debugging output stream.
 *
 * Returns:	
 *	 1 if debugging is successfully enabled, 0 if not.
 */
int debug_init(void)
{
	char *p = getenv("DEBUG_OUT");
	time_t now;

	if	(++_dbinits != 1)
	{
		return 1;
	}
		
	if	(p)
	{
		_dbfp = fopen(p, "a");
		if	(! _dbfp)
		{
			_dbfp = getSTDOUT();
			return 0;
		}

		time(&now);
		debug_print("**** Debug output: %s", ctime(&now));
	}
	else
		_dbfp = getSTDOUT();
	return 1;
}


/*
 * Function:  debug_shutdown
 *
 * Purpose:	
 *	 Shutdown the debugging system.  Primarily this means closing
 *   the debugging output stream.
 *
 * Returns:	
 *	 1 if debugging is successfully disabled, 0 if not.
 */
int debug_shutdown(void)
{
    	if	(--_dbinits != 0)
		return 0;
		
#ifndef	__linux__
	if	(_dbfp != stdout)
#endif
		if	(_dbfp != 0)
			fclose(_dbfp);
	_dbfp = 0;
	return 1;
}


/*
 * Function:  debug_print
 *
 * Purpose:	
 *	 Print out a debugging string to the debug stream.
 *
 * Returns:	
 *	 Return value of vfprintf.
 */
int debug_print(const char *format, ...)
{
	va_list	ap;
	int	ret;

	va_start (ap, format);

	if	(current_class) fprintf(_dbfp, "%s: ", current_class);
	ret = vfprintf (_dbfp, format, ap);
	fflush(_dbfp);
	va_end (ap);
	return ret;
}

int debug_continue(const char *format, ...)
{
	va_list	ap;
	int	ret;

	va_start (ap, format);

	ret = vfprintf (_dbfp, format, ap);
	fflush(_dbfp);
	va_end (ap);
	return ret;
}
